"""CLI to show HPC job information."""

import logging
import os
import sys
from pathlib import Path
from datetime import datetime, timedelta

import click
from prettytable import PrettyTable

from jade.exceptions import InvalidConfiguration
from jade.jobs.cluster import Cluster
from jade.hpc.common import HpcJobStats, HpcJobStatus
from jade.hpc.hpc_manager import HpcManager
from jade.loggers import setup_logging
from jade.models.submission_group import make_submission_group_lookup


logger = logging.getLogger(__name__)


def _check_output_dirs(_, __, output_dirs):
    if not output_dirs:
        print("output_dirs cannot be empty", file=sys.stderr)
        sys.exit(1)

    return output_dirs


@click.command()
@click.argument("output-dirs", nargs=-1, callback=_check_output_dirs)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def show_times(output_dirs, verbose):
    """Show the run times of all allocated jobs."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level)

    job_ids = []
    for output in output_dirs:
        path = Path(output)
        try:
            cluster, _ = Cluster.deserialize(path, deserialize_jobs=False)
        except InvalidConfiguration:
            print(f"{output} is not a JADE output directory used in cluster mode", file=sys.stderr)
            sys.exit(1)

        job_ids += [x.name.split("_")[2].replace(".e", "") for x in path.glob("*.e")]

    job_ids.sort(key=lambda x: int(x))
    groups = make_submission_group_lookup([cluster.config.submission_groups[0]])
    hpc_mgr = HpcManager(groups, output)

    total_duration = timedelta(seconds=0)
    table = PrettyTable()
    table.field_names = HpcJobStats._fields

    total_aus = 0
    if os.environ.get("NREL_CLUSTER") == "eagle":
        au_parser = get_nrel_eagle_aus
    else:
        au_parser = None

    for job_id in job_ids:
        stats = hpc_mgr.get_job_stats(job_id)
        if stats is None:
            continue
        duration = stats.end - stats.start
        if stats.state == HpcJobStatus.COMPLETE and isinstance(stats.end, datetime):
            total_duration += duration
        data = stats._asdict()
        data["state"] = data["state"].value
        if au_parser is not None:
            total_aus += au_parser(duration, stats.qos)
        table.add_row(data.values())

    print(table)
    print(f"\nTotal duration = {total_duration}")
    print("Total hours = {:.2f}".format(total_duration.total_seconds() / 3600))
    if au_parser is not None:
        print("Total AUs = {:.2f}".format(total_aus))


def get_nrel_eagle_aus(duration, qos):
    _duration = duration.total_seconds() / 3600
    if qos == "normal":
        val = _duration * 3.0
    elif qos == "standby":
        val = 0.0
    elif qos == "high":
        val = _duration * 3.0 * 2
    else:
        assert f"qos={qos} is not supported"

    return val


@click.command()
@click.argument("output-dirs", nargs=-1, callback=_check_output_dirs)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def list_active_ids(output_dirs, verbose):
    """List the HPC job IDs that are pending or running."""
    # TODO: add flag for only pending or only running
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level)

    job_ids = []
    for output in output_dirs:
        path = Path(output)
        try:
            cluster, _ = Cluster.deserialize(path, deserialize_jobs=True)
        except InvalidConfiguration:
            print(f"{output} is not a JADE output directory used in cluster mode", file=sys.stderr)
            sys.exit(1)

        if not cluster.is_complete():
            job_ids += list(cluster.iter_hpc_job_ids())

    job_ids.sort(key=lambda x: int(x))
    print(" ".join(job_ids))


@click.group()
def hpc_jobs():
    """Subcommands related to HPC jobs"""
    pass


hpc_jobs.add_command(list_active_ids)
hpc_jobs.add_command(show_times)



import logging
import shlex
import subprocess
import sys
import time

from jade.exceptions import ExecutionError, InvalidParameter
from jade.utils.timing_utils import timed_debug


logger = logging.getLogger(__name__)


@timed_debug
def run_command(
    cmd, output=None, cwd=None, num_retries=0, retry_delay_s=2.0, error_strings=None, **kwargs
):
    """Runs a command as a subprocess.

    Parameters
    ----------
    cmd : str
        command to run
    output : dict, default=None
        If a dict is passed then return stdout and stderr as keys.
    cwd : str, default=None
        Change the working directory to cwd before executing the process.
    num_retries : int, default=0
        Retry the command on failure this number of times.
        Return code and output are from the last command execution.
    retry_delay_s : float, default=2.0
        Number of seconds to delay in between retries.
    error_strings : None | str
        Skip retries for these known error strings. Requires output to be set.
    kwargs
        Keyword arguments to forward to the subprocess module.

    Returns
    -------
    int
        return code from system; usually zero is good, non-zero is error

    Caution: Capturing stdout and stderr in memory can be hazardous with
    long-running processes that output lots of text. In those cases consider
    running subprocess.Popen with stdout and/or stderr set to a pre-configured
    file descriptor.

    """
    if error_strings and output is None:
        raise InvalidParameter(f"output must be set if error_strings are passed")
    if error_strings is None:
        error_strings = []

    if not isinstance(cmd, str):
        # Handle pathlib.Path.
        cmd = str(cmd)
    logger.debug(cmd)
    # Disable posix if on Windows.
    command = shlex.split(cmd, posix="win" not in sys.platform)
    max_tries = num_retries + 1
    assert max_tries >= 1
    ret = None
    for i in range(max_tries):
        _output = {} if isinstance(output, dict) else None
        ret = _run_command(command, _output, cwd, **kwargs)
        if ret != 0 and num_retries > 0:
            if _output:
                if _should_exit_early(_output["stderr"], error_strings):
                    i = max_tries - 1
                else:
                    logger.warning(
                        "Command [%s] failed on iteration %s: %s: %s",
                        cmd,
                        i + 1,
                        ret,
                        _output["stderr"],
                    )
            else:
                logger.warning("Command [%s] failed on iteration %s: %s", cmd, i + 1, ret)
        if ret == 0 or i == max_tries - 1:
            if isinstance(output, dict):
                output.update(_output)
            break
        time.sleep(retry_delay_s)

    if ret != 0:
        logger.debug("Command [%s] failed: %s", cmd, ret)
        if output:
            logger.debug(output["stderr"])

    return ret


def _should_exit_early(std_err, error_strings):
    for err in error_strings:
        if err in std_err:
            return True
    return False


def _run_command(command, output, cwd, **kwargs):
    if output is not None:
        pipe = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd, **kwargs
        )
        out, err = pipe.communicate()
        output["stdout"] = out.decode("utf-8")
        output["stderr"] = err.decode("utf-8")
        ret = pipe.returncode
    else:
        ret = subprocess.call(command, cwd=cwd, **kwargs)

    return ret


def check_run_command(*args, **kwargs):
    """Same as run_command except that it raises an exception on failure.

    Raises
    ------
    ExecutionError
        Raised if the command returns a non-zero return code.

    """
    ret = run_command(*args, **kwargs)
    if ret != 0:
        raise ExecutionError(f"command returned error code: {ret}")



"""CLI to run hosting capacity analysis on results for a feeder."""

import logging
import os
import sys

import click

from jade.common import OUTPUT_DIR
from jade.events import StructuredErrorLogEvent, EVENT_CATEGORY_ERROR, EVENT_NAME_UNHANDLED_ERROR
from jade.loggers import log_event, setup_logging, setup_event_logging
from jade.jobs.job_post_process import JobPostProcess
from jade.utils.utils import get_cli_string, load_data
from jade.exceptions import InvalidExtension
from jade.extensions.registry import Registry, ExtensionClassType


@click.argument("extension")
@click.option(
    "-n",
    "--name",
    required=True,
    type=str,
    help="The name of the job that needs to run.",
)
@click.option("-o", "--output", default=OUTPUT_DIR, show_default=True, help="Output directory.")
@click.option("--config-file", required=True, help="Job configuration file")
@click.option(
    "-f",
    "--output-format",
    default="csv",
    show_default=True,
    help="Output format for data (csv or json).",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
@click.command()
def run(extension, **kwargs):
    """Runs individual job."""
    registry = Registry()
    if not registry.is_registered(extension):
        raise InvalidExtension(f"Extension '{extension}' is not registered.")

    # Parse Argument
    config_file = kwargs["config_file"]
    name = kwargs["name"]
    output = kwargs["output"]
    output_format = kwargs["output_format"]
    verbose = kwargs["verbose"]
    level = logging.DEBUG if verbose else logging.INFO

    # Create directory for current job
    job_dir = os.path.join(output, name)
    os.makedirs(job_dir, exist_ok=True)
    # Structural logging setup
    event_file = os.path.join(job_dir, "events.log")
    setup_event_logging(event_file, mode="a")

    # General logging setup
    log_file = os.path.join(job_dir, "run.log")
    general_logger = setup_logging(
        extension,
        log_file,
        console_level=logging.ERROR,
        file_level=level,
    )
    general_logger.info(get_cli_string())

    # Create config for run
    try:
        cli = registry.get_extension_class(extension, ExtensionClassType.CLI)
        ret = cli.run(config_file, name, output, output_format, verbose)
    except Exception as err:
        msg = f"unexpected exception in run '{extension}' job={name} - {err}"
        general_logger.exception(msg)
        event = StructuredErrorLogEvent(
            source=name,
            category=EVENT_CATEGORY_ERROR,
            name=EVENT_NAME_UNHANDLED_ERROR,
            message=msg,
        )
        log_event(event)
        ret = 1

    if ret == 0:
        try:
            config = load_data(config_file)
            if "job_post_process_config" in config.keys():
                post_process = JobPostProcess(
                    module_name=config["job_post_process_config"]["module"],
                    class_name=config["job_post_process_config"]["class"],
                    data=config["job_post_process_config"]["data"],
                    job_name=name,
                    output=output,
                )
                post_process.run(config_file=config_file, output=output)
        except Exception as err:
            msg = f"unexpected exception in post-process '{extension}' job={name} - {err}"
            general_logger.exception(msg)
            event = StructuredErrorLogEvent(
                source=name,
                category=EVENT_CATEGORY_ERROR,
                name=EVENT_NAME_UNHANDLED_ERROR,
                message=msg,
            )
            log_event(event)
            ret = 1

    sys.exit(ret)



import logging
import os
import shutil
import stat
import subprocess
import sys
import time
from pathlib import Path

import click

from jade.common import HPC_CONFIG_FILE, OUTPUT_DIR
from jade.loggers import setup_logging
from jade.models import HpcConfig, get_model_defaults
from jade.models.spark import SparkConfigModel, SparkContainerModel
from jade.utils.utils import dump_data, load_data


SPARK_MODEL_DEFAULTS = get_model_defaults(SparkConfigModel)
DYNAMIC_ALLOCATION_SETTINGS = """spark.dynamicAllocation.enabled true
spark.dynamicAllocation.shuffleTracking.enabled true
spark.shuffle.service.enabled true"""
GPU_DISCOVERY_SCRIPT = "/opt/sparkRapidsPlugin/getGpusResources.sh"


def _replace_tag(values, filename):
    text = filename.read_text()
    for tag, value in values:
        text = text.replace(f"<{tag}>", value)
    filename.write_text(text)


def _gpu_cb(_, __, val):
    if val is None:
        return val
    val = val.lower()
    if val == "true":
        return True
    if val == "false":
        return False
    raise click.BadParameter(f"Supported values for 'gpu': true or false")


@click.command()
@click.option(
    "-C",
    "--collect-worker-logs",
    default=False,
    is_flag=True,
    show_default=True,
    help="Collect logs from worker processes.",
)
@click.option(
    "-c",
    "--container-path",
    type=str,
    required=True,
    help="Path to container that can run Spark",
)
@click.option(
    "--dynamic-allocation/--no-dynamic-allocation",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable Spark dynamic resource allocation.",
)
@click.option(
    "-g",
    "--gpu",
    default=None,
    callback=_gpu_cb,
    help="Possible values: not set (default), 'true', 'false'. If not set, create GPU-ready "
    "config if hpc_config.toml specifies GPUs.",
)
@click.option(
    "-h",
    "--hpc-config",
    type=click.Path(exists=True),
    default=HPC_CONFIG_FILE,
    show_default=True,
    help="HPC config file to be used. Defines number of nodes.",
)
@click.option(
    "-m",
    "--master-node-memory-overhead-gb",
    default=SPARK_MODEL_DEFAULTS["master_node_memory_overhead_gb"],
    show_default=True,
    help="Memory overhead for Spark master processes",
)
@click.option(
    "-n",
    "--node-memory-overhead-gb",
    default=SPARK_MODEL_DEFAULTS["node_memory_overhead_gb"],
    show_default=True,
    help="Memory overhead for node operating system and existing applications",
)
@click.option(
    "-r",
    "--run-user-script-inside-container",
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
    help="Run the user script inside the container.",
)
@click.option(
    "-S",
    "--spark-dir",
    type=str,
    default="spark",
    show_default=True,
    help="Spark configuration directory to create.",
)
@click.option(
    "-s",
    "--shuffle-partition-multiplier",
    type=int,
    default=1,
    show_default=True,
    help="Set spark.sql.shuffle.partitions to total_cores multiplied by this value.",
)
@click.option(
    "-U",
    "--use-tmpfs-for-scratch",
    is_flag=True,
    default=False,
    show_default=True,
    help="Use tmpfs on node for scratch space.",
)
@click.option(
    "-a",
    "--alt-scratch",
    default=None,
    show_default=True,
    help="Use this alternative directory for scratch space.",
)
@click.option(
    "-u",
    "--update-config-file",
    type=str,
    default=None,
    help="Update all jobs in this config file with this Spark configuration.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable verbose log output.",
)
@click.option(
    "-W",
    "--worker-memory-gb",
    default=SPARK_MODEL_DEFAULTS["worker_memory_gb"],
    show_default=True,
    help="If 0, give all node memory minus overhead to worker.",
)
@click.option(
    "-f",
    "--force",
    default=False,
    is_flag=True,
    show_default=True,
    help="Overwrite the spark configuration directory if it already exists.",
)
def config(
    collect_worker_logs,
    container_path,
    dynamic_allocation,
    gpu,
    hpc_config,
    master_node_memory_overhead_gb,
    node_memory_overhead_gb,
    run_user_script_inside_container,
    spark_dir,
    shuffle_partition_multiplier,
    update_config_file,
    use_tmpfs_for_scratch,
    alt_scratch,
    verbose,
    worker_memory_gb,
    force,
):
    """Create a Spark configuration to use for running a job on a Spark cluster."""
    level = logging.DEBUG if verbose else logging.WARNING
    setup_logging("config_spark", None, console_level=level)
    spark_dir = Path(spark_dir)
    if spark_dir.exists():
        if force:
            shutil.rmtree(spark_dir)
        else:
            print(
                f"The directory '{spark_dir}' already exists. Use a different name or pass --force to overwrite.",
                file=sys.stderr,
            )
            sys.exit(1)
    spark_dir.mkdir(parents=True)

    if use_tmpfs_for_scratch and alt is not None:
        print("use_tmpfs_for_scratch and alt_scratch cannot both be set", file=sys.stderr)
        sys.exit(1)

    hpc_config_data = HpcConfig.load(hpc_config)
    nodes = getattr(hpc_config_data.hpc, "nodes", None)
    if nodes is None:
        print(f"hpc_type={hpc_config_data.hpc_type} doesn't have a nodes field", file=sys.stderr)
        sys.exit(1)
    mem = getattr(hpc_config_data.hpc, "mem", None)
    if mem is None:
        executor_mem_gb = 11
        print(f"Use default per-executor memory of {executor_mem_gb}G")
    else:
        num_executors = 7
        if not mem.endswith("G"):
            raise Exception(f"This feature only supports HPC memory requirements ending with 'G'")
        per_node_mem_gb = int(mem[:-1])
        if use_tmpfs_for_scratch:
            per_node_mem_gb //= 2
        overhead = master_node_memory_overhead_gb - node_memory_overhead_gb
        executor_mem_gb = (per_node_mem_gb - overhead) // num_executors
        print(f"Use custom per-executor memory of {executor_mem_gb}G based on per-node {mem}")

    for dirname in ("bin", "conf"):
        src_path = Path(os.path.dirname(__file__)).parent / "spark" / dirname
        dst_path = spark_dir / dirname
        if not dst_path.exists():
            dst_path.mkdir()
        for filename in src_path.iterdir():
            shutil.copyfile(filename, dst_path / filename.name)

    use_gpus = _should_use_gpus(hpc_config_data, gpu)

    with open(spark_dir / "conf" / "spark-defaults.conf", "a") as f_out:
        f_out.write("\n")
        f_out.write(f"spark.executor.memory {executor_mem_gb}G\n")
        # Online documentation says this value should correlate with the number of cores in the
        # cluster. Some sources say 1 per core, others say 2 or 4 per core. Depends on use case.
        # This should be a reasonable default for users, who can customize dynamically.
        params = ["spark.sql.shuffle.partitions"]
        # Some sources say that we should set spark.default.parallelism to the same value,
        # others say it doesn't work. Experiments showed harmful effects if dynamic allocation
        # was enabled with a custom value.
        for param in params:
            f_out.write(param)
            f_out.write(" ")
            f_out.write(str(nodes * 35 * shuffle_partition_multiplier))
            f_out.write("\n")

        if dynamic_allocation:
            f_out.write("\n")
            f_out.write(DYNAMIC_ALLOCATION_SETTINGS)
            f_out.write("\n")

        if use_gpus:
            src_path = (
                Path(os.path.dirname(__file__)).parent / "spark" / "conf" / "resourcesFile.json"
            )
            resources_file = spark_dir / "conf" / "resourcesFile.json"
            shutil.copyfile(src_path, resources_file)
            f_out.write(
                "spark.worker.resource.gpu.discoveryScript /opt/sparkRapidsPlugin/getGpusResources.sh\n"
            )
            f_out.write(f"spark.worker.resourcesFile {resources_file}\n")

    if use_gpus:
        filename = spark_dir / "conf" / "spark-env.sh"
        with open(filename, "a") as f_out:
            num_gpus = hpc_config_data.get_num_gpus() or 2
            f_out.write(
                f'SPARK_WORKER_OPTS="-Dspark.worker.resource.gpu.amount={num_gpus} '
                f'-Dspark.worker.resource.gpu.discoveryScript={GPU_DISCOVERY_SCRIPT}"\n'
            )

    replacement_values = [
        ("SPARK_DIR", str(spark_dir)),
        ("CONTAINER_PATH", container_path),
    ]
    for name in ("run_spark_script_wrapper.sh", "run_user_script_wrapper.sh"):
        filename = spark_dir / "bin" / name
        _replace_tag(replacement_values, filename)
        st = os.stat(filename)
        os.chmod(filename, st.st_mode | stat.S_IEXEC)
        print(f"Assigned paths in {filename}")

    scripts = [spark_dir / "conf" / "spark-env.sh"] + list((spark_dir / "bin").glob("*.sh"))
    for script in scripts:
        st = os.stat(script)
        os.chmod(script, st.st_mode | stat.S_IEXEC)

    print(
        f"Created Spark configuration in {spark_dir.absolute()} for a {nodes}-node cluster. "
        f"GPUs={use_gpus}"
    )

    spark_config = SparkConfigModel(
        collect_worker_logs=collect_worker_logs,
        conf_dir=str(spark_dir),
        container=SparkContainerModel(path=container_path),
        enabled=True,
        master_node_memory_overhead_gb=master_node_memory_overhead_gb,
        node_memory_overhead_gb=node_memory_overhead_gb,
        run_user_script_inside_container=run_user_script_inside_container,
        use_tmpfs_for_scratch=use_tmpfs_for_scratch,
        alt_scratch=alt_scratch,
        worker_memory_gb=worker_memory_gb,
    )

    if update_config_file is not None:
        if not Path(update_config_file).exists():
            print(f"'update_config_file={update_config_file} does not exist", file=sys.stderr)
            sys.exit(1)
        config = load_data(update_config_file)
        for job in config["jobs"]:
            job["spark_config"] = spark_config.dict()
        dump_data(config, update_config_file, indent=2)
        print(f"Updated jobs in {update_config_file} with this Spark configuration.")
    else:
        print(
            "\nAdd and customize this JSON object to the 'spark_config' field for each Spark "
            "job in your config.json file:\n"
        )
        print(spark_config.json(indent=2))


def _should_use_gpus(hpc_config, gpu):
    hc_gpus = getattr(hpc_config.hpc, "gres", None)
    if gpu is None:
        use_gpus = hc_gpus is not None
    elif gpu:
        use_gpus = True
    else:
        use_gpus = False
    return use_gpus


@click.command()
@click.argument("compute-nodes", nargs=-1)
@click.option(
    "-C",
    "--container",
    required=True,
    help="Path to Spark container",
    type=click.Path(exists=True),
    callback=lambda _, __, x: Path(x),
)
@click.option(
    "-c",
    "--spark-conf",
    required=True,
    help="Spark conf directory",
    type=click.Path(exists=True),
    callback=lambda _, __, x: Path(x),
)
@click.option(
    "-s",
    "--script",
    help="Script to run on cluster. If None, sleep indefinitely.",
)
@click.option(
    "-o",
    "--output",
    default=OUTPUT_DIR,
    help="Output directory",
    callback=lambda _, __, x: Path(x),
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    show_default=True,
    help="Overwrite output directory if it exists.",
)
def start_cluster(compute_nodes, container, spark_conf, script, output, force):
    conda_env = os.environ.get("CONDA_DEFAULT_ENV")
    if conda_env is None:
        print(f"Did not detect conda environment. Exiting.", file=sys.stderr)
        sys.exit(1)
    if not compute_nodes:
        print("compute-nodes cannot be empty", file=sys.stderr)
        sys.exit(1)
    if output.exists():
        if force:
            shutil.rmtree(output)
        else:
            print(f"{output} already exists. Choose a different path or set --force to overwrite.")
            sys.exit(1)
    output.mkdir()
    master_node = compute_nodes[0]
    worker_nodes = []
    if len(compute_nodes) > 1:
        worker_nodes = compute_nodes[1:]

    _start_spark_master(master_node, container, spark_conf, script, output, conda_env)
    for worker in worker_nodes:
        _start_spark_worker(worker, master_node, container, spark_conf, output, conda_env)


def _start_spark_master(node_name, container, spark_conf, script, output_dir, conda_env):
    container = container.absolute()
    spark_conf = spark_conf.absolute()
    output_dir = output_dir.absolute()
    cwd = os.getcwd()

    script_option = "" if script is None else f"-s {os.path.abspath(script)}"
    text = f"""#!/bin/bash
ssh {node_name} "conda activate {conda_env}
cd {cwd}
jade-internal start-spark-cluster-master {container} {spark_conf} {output_dir} {script_option}"
"""
    spark_script = Path("/tmp/create_spark_master.sh")
    spark_script.write_text(text)
    try:
        subprocess.Popen(["bash", spark_script])
        time.sleep(2)
    finally:
        spark_script.unlink()


def _start_spark_worker(
    node_name, cluster_master_node_name, container, spark_conf, output_dir, conda_env
):
    container = container.absolute()
    spark_conf = spark_conf.absolute()
    output_dir = output_dir.absolute()
    cwd = os.getcwd()
    text = f"""#!/bin/bash
ssh {node_name} "conda activate {conda_env}
cd {cwd}
jade-internal start-spark-cluster-worker {cluster_master_node_name} {container} {spark_conf} {output_dir}"
"""
    script = Path("/tmp/create_spark_cluster_worker.sh")
    script.write_text(text)
    try:
        subprocess.Popen(["bash", script])
        time.sleep(2)
    finally:
        script.unlink()


@click.group()
def spark():
    """Spark commands"""


spark.add_command(config)
spark.add_command(start_cluster)



"""Utility functions for the jade package."""

from datetime import datetime, date
from pathlib import PosixPath, WindowsPath
from typing import Union
import enum
import functools
import gzip
import logging
import json
import os
import re
import shutil
import stat
import sys
from dateutil.parser import parse

import toml
from pydantic import BaseModel

from jade.exceptions import InvalidParameter
from jade.utils.timing_utils import timed_debug


MAX_PATH_LENGTH = 255

logger = logging.getLogger(__name__)


def create_chunks(items, size):
    """Returns a generator dividing items into chunks.

    items : list
    batch_size : int

    Returns
    -------
    generator

    """
    for i in range(0, len(items), size):
        yield items[i : i + size]


def create_script(filename, text, executable=True):
    """Creates a script with the given text.

    Parameters
    ----------
    text : str
        body of script
    filename : str
        file to create
    executable : bool
        if True, set as executable

    """
    # Permissions issues occur when trying to overwrite and then make
    # executable another user's file.
    if os.path.exists(filename):
        os.remove(filename)

    with open(filename, "w") as f_out:
        logger.debug("Writing %s", filename)
        f_out.write(text)

    if executable:
        curstat = os.stat(filename)
        os.chmod(filename, curstat.st_mode | stat.S_IEXEC)


def make_directory_read_only(directory):
    """Set all files in the directory to be read-only.

    Parameters
    ----------
    directory : str

    """
    for filename in os.listdir(directory):
        make_file_read_only(os.path.join(directory, filename))

    logger.debug("Made all files in %s read-only", directory)


def make_file_read_only(filename):
    """Set the file to be read-only.

    Parameters
    ----------
    filename : str

    """
    os.chmod(filename, 0o444)
    logger.debug("Set %s as read-only", filename)


def _get_module_from_extension(filename, **kwargs):
    ext = os.path.splitext(filename)[1].lower()
    if ext == ".json":
        mod = json
    elif ext == ".toml":
        mod = toml
    # elif ext in (".yml", ".yaml"):
    #    mod = yaml
    elif "mod" in kwargs:
        mod = kwargs["mod"]
    else:
        raise InvalidParameter(f"Unsupported extension {filename}")

    return mod


@timed_debug
def dump_data(data, filename, **kwargs):
    """Dump data to the filename.
    Supports JSON, TOML, YAML, or custom via kwargs.

    Parameters
    ----------
    data : dict
        data to dump
    filename : str
        file to create or overwrite

    """
    mod = _get_module_from_extension(filename, **kwargs)
    with open(filename, "w") as f_out:
        mod.dump(data, f_out, **kwargs)

    logger.debug("Dumped data to %s", filename)


@timed_debug
def load_data(filename, **kwargs):
    """Load data from the file.
    Supports JSON, TOML, or custom via kwargs.
    YAML support could easily be added.

    Parameters
    ----------
    filename : str

    Returns
    -------
    dict

    """
    # TODO:  YAMLLoadWarning: calling yaml.load() without Loader=... is deprecated,
    #  as the default Loader is unsafe. Please read https://msg.pyyaml.org/load for full details.
    mod = _get_module_from_extension(filename, **kwargs)
    with open(filename) as f_in:
        try:
            data = mod.load(f_in)
        except Exception:
            logger.exception(f"Failed to load {filename}")
            raise

    logger.debug("Loaded data from %s", filename)
    return data


def aggregate_data_from_files(directory, end_substring, **kwargs):
    """Aggregate objects from files in directory matching end_substring.
    Refer to :func:`~jade.utils.utils.load_data` for supported file formats.

    Parameters
    ----------
    directory : str
    substring : str

    Returns
    -------
    list of dict

    """
    data = []
    for filename in os.listdir(directory):
        if filename.endswith(end_substring):
            path = os.path.join(directory, filename)
            data.append(load_data(path, **kwargs))

    return data


def rmtree(path):
    """Deletes the directory tree if it exists.

    Parameters
    ----------
    path : str

    """
    if os.path.exists(path):
        shutil.rmtree(path)

    logger.debug("Deleted %s", path)


def modify_file(filename, line_func, *args, **kwargs):
    """Modifies a file by running a function on each line.

    Parameters
    ----------
    filename : str
    line_func : callable
        Should return the line to write to the modified file.

    """
    tmp = filename + ".tmp"
    assert not os.path.exists(tmp), f"filename={filename} tmp={tmp}"

    with open(filename) as f_in:
        with open(tmp, "w") as f_out:
            for line in f_in:
                line = line_func(line, *args, **kwargs)
                f_out.write(line)

    shutil.move(tmp, filename)


def get_cli_string():
    """Return the command-line arguments issued.

    Returns
    -------
    str

    """
    return os.path.basename(sys.argv[0]) + " " + " ".join(sys.argv[1:])


def handle_file_not_found(func):
    """Decorator to catch FileNotFoundError exceptions."""

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except FileNotFoundError:
            msg = "one or more input parameters do not exist"
            logger.debug(msg, exc_info=True)
            raise InvalidParameter("{}: {}".format(msg, args[1:]))

        return result

    return wrapped


def handle_key_error(func):
    """Decorator to catch KeyError exceptions that happen because the user
    performs invalid actions."""

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except KeyError as err:
            msg = "invalid parameter: {}".format(err)
            logger.debug(msg, exc_info=True)
            raise InvalidParameter(msg)

        return result

    return wrapped


def decompress_file(filename):
    """Decompress a file.

    Parameters
    ----------
    filename : str

    Returns
    -------
    str
        Returns the new filename.

    """
    assert os.path.splitext(filename)[1] == ".gz"

    new_filename = filename[:-3]
    with open(new_filename, "wb") as f_out:
        with gzip.open(filename, "rb") as f_in:
            shutil.copyfileobj(f_in, f_out)

    os.remove(filename)
    logger.debug("Decompressed %s", new_filename)
    return new_filename


def get_directory_size_bytes(directory, recursive=True):
    """Return the total space consumed by all files in directory.

    Parameters
    ----------
    directory : str
    recursive : bool

    Returns
    -------
    int
        size in bytes

    """
    total = 0
    for dirpath, _, filenames in os.walk(directory):
        for filename in filenames:
            total += os.stat(os.path.join(dirpath, filename)).st_size
        if not recursive:
            break

    return total


def get_filenames_in_path(directory, filename, is_regex=False):
    """Return all files in directory matching filename, searching recursively.

    Parameters
    ----------
    directory : str
    ext : str
        file extension, ex: .log
    is_regex : bool
        Treat filename as regular expression.

    Returns
    -------
    generator

    """
    for dirpath, _, filenames in os.walk(directory):
        for _filename in filenames:
            matched = False
            if is_regex:
                matched = filename.search(_filename)
            else:
                matched = _filename == filename
            if matched:
                yield os.path.join(dirpath, _filename)


def get_filenames_by_ext(directory, ext):
    """Return filenames in directory, recursively, with file extension.

    Parameters
    ----------
    directory : str
    ext : str
        file extension, ex: .log

    Returns
    -------
    generator

    """
    for dirpath, _, filenames in os.walk(directory):
        for filename in filenames:
            if ext in filename:
                yield os.path.join(dirpath, filename)


def interpret_datetime(timestamp):
    """Return a datetime object from a timestamp string.

    Parameters
    ----------
    timestamp : str

    Returns
    -------
    datetime.datetime

    """
    formats = (
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%d %H:%M:%S.%f",
        "%Y-%m-%dT%H:%M:%SZ",
        "%Y-%m-%dT%H:%M:%S.%fZ",
        "%Y-%m-%d_%H:%M:%S.%f",
        "%Y-%m-%d_%H-%M-%S-%f",
    )

    for i, fmt in enumerate(formats):
        try:
            return datetime.strptime(timestamp, fmt)
        except ValueError:
            if i == len(formats) - 1:
                raise
            continue


def standardize_timestamp(timestamp: Union[str, datetime]) -> str:
    """Validate string timestamp and output standard format."""
    stdfmt = "%Y-%m-%dT%H:%M:%S.%f"
    if isinstance(timestamp, datetime):
        return timestamp.strftime(stdfmt)

    dt = None
    formats = ("%Y-%m-%d_%H:%M:%S.%f", "%Y-%m-%d_%H-%M-%S-%f")
    for fmt in formats:
        try:
            dt = datetime.strptime(timestamp, fmt)
        except ValueError:
            continue

    if not dt:
        dt = parse(timestamp)

    return dt.strftime(stdfmt)


def rotate_filenames(directory, ext):
    """Rotates filenames in directory, recursively, to .1, .2, .etc.

    Parameters
    ----------
    directory : str
    ext : str
        file extension, ex: .log

    """
    regex_has_num = re.compile(r"(\d+)(?:\.[a-zA-Z]+)?$")

    for dirpath, _, filenames in os.walk(directory):
        files = []
        for filename in filter(lambda x: ext in x, filenames):
            new_names = set()
            match = regex_has_num.search(filename)
            if match:
                cur = int(match.group(1))
                new = cur + 1
                new_name = filename.replace(f"{ext}.{cur}", f"{ext}.{new}")
            else:
                cur = 0
                if not filename.endswith(ext):
                    # such as compressed files
                    split_ext = os.path.splitext(filename)
                    new_name = split_ext[0] + ".1" + split_ext[1]
                else:
                    new_name = filename + ".1"
            assert new_name not in new_names
            new_names.add(new_name)
            files.append((filename, cur, new_name))

        files.sort(key=lambda x: x[1], reverse=True)
        for filename in files:
            old = os.path.join(dirpath, filename[0])
            new = os.path.join(dirpath, filename[2])
            os.rename(old, new)
            logger.info("Renamed %s to %s", old, new)


def check_filename(name):
    """
    Validates that a name is valid for use as a filename or directory.
    Valid characters:  letters, numbers, underscore, hyphen, period

    Parameters
    ----------
    string: str,
        A given string.

    Raises
    ------
    InvalidParameter
        Raised if the name contains illegal characters or is too long.

    """
    if not re.search(r"^[\w\.-]+$", name):
        raise InvalidParameter(f"{name} contains illegal characters.")

    if len(name) > MAX_PATH_LENGTH:
        raise InvalidParameter(f"length of {name} is greater than the limit of {MAX_PATH_LENGTH}.")


def output_to_file(data, filename=None, stream=sys.stdout, indent=2):
    if filename is None and stream is None:
        raise InvalidParameter("must set either filename or stream")

    if filename is not None:
        ext = os.path.splitext(filename)[1]
        if ext not in (".json", ".toml"):
            raise InvalidParameter("Only .json and .toml are supported")

        with open(filename, "w") as f_out:
            _write_file(data, f_out, fmt=ext)
    else:
        _write_file(data)

    logger.info("Dumped configuration to %s", filename)


def _write_file(data, stream=sys.stdout, fmt=".json", indent=2):
    # Note: the default is JSON here because parsing 100 MB .toml files
    # is an order of magnitude slower.
    if fmt == ".json":
        json.dump(data, stream, indent=indent)
    elif fmt == ".toml":
        toml.dump(data, stream)
    else:
        assert False, fmt


class ExtendedJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, enum.Enum):
            return obj.value

        if isinstance(obj, PosixPath) or isinstance(obj, WindowsPath):
            return str(obj)

        if isinstance(obj, (datetime, date)):
            return standardize_timestamp(obj)

        if isinstance(obj, set):
            return list(obj)

        if isinstance(obj, BaseModel):
            return obj.dict()

        return json.JSONEncoder.default(self, obj)



"""Implements a container for jobs represented by a key."""

from collections import OrderedDict
import logging

from jade.exceptions import InvalidParameter
from jade.jobs.job_container_interface import JobContainerInterface
from jade.utils.utils import check_filename


logger = logging.getLogger(__name__)


class JobContainerByKey(JobContainerInterface):
    """Stores jobs by key which is a namedtuple."""

    def __init__(self):
        # collections.namedtuple: JobParametersInterface
        self._jobs = OrderedDict()

    def __iter__(self):
        for job in self._jobs.values():
            yield job

    def __len__(self):
        return len(self._jobs)

    @staticmethod
    def _get_key(job=None, key=None):
        if key is None and job is None:
            raise InvalidParameter("either key or job must be passed")
        if key is not None and job is not None:
            raise InvalidParameter("only one of key and job can be " "passed")
        if key is None:
            key = job.name

        return key

    def add_job(self, job, key=None):
        if key is None:
            key = job.name

        if key in self._jobs:
            raise InvalidParameter(f"key={key} is already stored")

        check_filename(key)

        self._jobs[key] = job
        logger.debug("Added job %s", key)

    def clear(self):
        self._jobs.clear()
        logger.debug("Cleared all jobs.")

    def remove_job(self, job=None, key=None):
        key = self._get_key(job=job, key=key)
        self._jobs.pop(key)
        logger.info("Removed job with key=%s", key)

    def get_job(self, name):
        for job in self:
            if job.name == name:
                return job

        raise InvalidParameter(f"job {name} not found")

    def get_job_by_key(self, key):
        job = self._jobs.get(key)
        if job is None:
            raise InvalidParameter(f"job key={key} not found")

        return job

    def get_jobs(self, sort=False):
        if sort:
            keys = list(self._jobs.keys())
            keys.sort()
            return [self._jobs[x] for x in keys]

        return list(self)

    def shuffle(self):
        keys = list(self._jobs.keys())
        random.shuffle(keys)
        new_jobs = {}
        for key in keys:
            new_jobs[key] = self._jobs.pop(key)
        self._jobs = new_jobs
